"""The lower-level records API."""

import re
from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    ClassVar,
    Iterable,
    List,
    Optional,
    Pattern,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from attr.validators import optional
from attrs import define, field

from netsgiro import RecordType, ServiceCode
from netsgiro.converters import (
    fixed_len_str,
    stripped_newlines,
    to_assignment_type,
    to_avtalegiro_registration_type,
    to_bool,
    to_date,
    to_date_or_none,
    to_int_or_none,
    to_record_type,
    to_safe_str_or_none,
    to_service_code,
    to_transaction_type,
)
from netsgiro.validators import str_of_length, str_of_max_length

if TYPE_CHECKING:
    import datetime

    from netsgiro.enums import AssignmentType, AvtaleGiroRegistrationType, TransactionType

__all__: List[str] = [
    'TransmissionStart',
    'TransmissionEnd',
    'AssignmentStart',
    'AssignmentEnd',
    'TransactionAmountItem1',
    'TransactionAmountItem2',
    'TransactionAmountItem3',
    'TransactionSpecification',
    'AvtaleGiroAgreement',
    'parse',
]

R = TypeVar('R', bound='Record')


@define
class Record(ABC):
    """Record base class."""

    _PATTERNS: ClassVar[List[Pattern]]
    RECORD_TYPE: ClassVar[RecordType]

    service_code: ServiceCode = field(converter=to_service_code)

    @classmethod
    def from_string(cls: Type[R], line: str) -> R:
        """Parse OCR string into a record object."""
        for pattern in cls._PATTERNS:
            matches = pattern.match(line)
            if matches is not None:
                return cls(**matches.groupdict())

        raise ValueError(f'{line!r} did not match {cls.__name__} record formats')

    @abstractmethod
    def to_ocr(self) -> str:
        """Get record as OCR string."""


@define
class TransmissionStart(Record):
    """TransmissionStart is the first record in every OCR file.

    A file can only contain a single transmission.

    Each transmission can contain any number of assignments.
    """

    transmission_number: str = field(validator=str_of_length(7))
    data_transmitter: str = field(validator=str_of_length(8))
    data_recipient: str = field(validator=str_of_length(8))

    RECORD_TYPE: ClassVar[RecordType] = RecordType.TRANSMISSION_START
    _PATTERNS: ClassVar[List[Pattern]] = [
        re.compile(
            r'''
            ^
            NY      # Format code
            (?P<service_code>00)
            00      # Transmission type, always 00
            10      # Record type

            (?P<data_transmitter>\d{8})
            (?P<transmission_number>\d{7})
            (?P<data_recipient>\d{8})

            0{49}   # Padding
            $
            ''',
            re.VERBOSE,
        )
    ]

    def to_ocr(self) -> str:
        """Get record as OCR string."""
        return (
            f'NY000010{self.data_transmitter:8}{self.transmission_number:7}{self.data_recipient:8}'
            + ('0' * 49)
        )


@define
class TransmissionEnd(Record):
    """TransmissionEnd is the first record in every OCR file."""

    num_transactions: int = field(converter=int)
    num_records: int = field(converter=int)
    total_amount: int = field(converter=int)
    nets_date: 'datetime.date' = field(converter=to_date)

    RECORD_TYPE: ClassVar[RecordType] = RecordType.TRANSMISSION_END
    _PATTERNS: ClassVar[List[Pattern]] = [
        re.compile(
            r'''
            ^
            NY      # Format code
            (?P<service_code>00)
            00      # Transmission type, always 00
            89      # Record type

            (?P<num_transactions>\d{8})
            (?P<num_records>\d{8})
            (?P<total_amount>\d{17})
            (?P<nets_date>\d{6})

            0{33}   # Filler
            $
            ''',
            re.VERBOSE,
        )
    ]

    def to_ocr(self) -> str:
        """Get record as OCR string."""
        return (
            'NY000089'
            f'{self.num_transactions:08d}'
            f'{self.num_records:08d}'
            f'{self.total_amount:017d}'
            f'{self.nets_date:%d%m%y}'
            + ('0' * 33)
        )


@define
class AssignmentStart(Record):
    """AssignmentStart is the first record of an assignment.

    Each assignment can contain any number of transactions.
    """

    assignment_type: 'AssignmentType' = field(converter=to_assignment_type)
    assignment_number: str = field(validator=str_of_length(7))
    assignment_account: str = field(validator=str_of_length(11))

    # Only for assignment_type == AssignmentType.TRANSACTIONS
    agreement_id: Optional[str] = field(default=None, validator=optional(str_of_length(9)))

    RECORD_TYPE: ClassVar[RecordType] = RecordType.ASSIGNMENT_START
    _PATTERNS: ClassVar[List[Pattern]] = [
        re.compile(
            r'''
            ^
            NY      # Format code
            (?P<service_code>(09|21))
            (?P<assignment_type>00)
            20      # Record type

            (?P<agreement_id>\d{9})
            (?P<assignment_number>\d{7})
            (?P<assignment_account>\d{11})

            0{45}   # Filler
            $
            ''',
            re.VERBOSE,
        ),
        re.compile(
            r'''
            ^
            NY      # Format code
            (?P<service_code>21)
            (?P<assignment_type>24)
            20      # Record type

            0{9}    # Filler

            (?P<assignment_number>\d{7})
            (?P<assignment_account>\d{11})

            0{45}   # Filler
            $
            ''',
            re.VERBOSE,
        ),
        re.compile(
            r'''
            ^
            NY      # Format code
            (?P<service_code>21)
            (?P<assignment_type>36)
            20      # Record type

            0{9}    # Filler

            (?P<assignment_number>\d{7})
            (?P<assignment_account>\d{11})

            0{45}   # Filler
            $
            ''',
            re.VERBOSE,
        ),
    ]

    def to_ocr(self) -> str:
        """Get record as OCR string."""
        return (
            f'NY{self.service_code:02d}{self.assignment_type:02d}20'
            + (self.agreement_id and f'{self.agreement_id:9}' or ('0' * 9))
            + f'{self.assignment_number:7}{self.assignment_account:11}'
            + ('0' * 45)
        )


@define
class AssignmentEnd(Record):
    """AssignmentEnd is the last record of an assignment."""

    assignment_type: 'AssignmentType' = field(converter=to_assignment_type)
    num_transactions: int = field(converter=int)
    num_records: int = field(converter=int)

    # Only for transactions and cancellations
    total_amount: Optional[int] = field(default=None, converter=to_int_or_none)
    nets_date_1: Optional['datetime.date'] = field(default=None, converter=to_date_or_none)
    nets_date_2: Optional['datetime.date'] = field(default=None, converter=to_date_or_none)
    nets_date_3: Optional['datetime.date'] = field(default=None, converter=to_date_or_none)

    RECORD_TYPE: ClassVar[RecordType] = RecordType.ASSIGNMENT_END
    _PATTERNS: ClassVar[List[Pattern]] = [
        re.compile(
            r'''
            ^
            NY      # Format code
            (?P<service_code>(09|21))
            (?P<assignment_type>00)     # Transactions / payment requests
            88      # Record type

            (?P<num_transactions>\d{8})
            (?P<num_records>\d{8})
            (?P<total_amount>\d{17})
            (?P<nets_date_1>\d{6})
            (?P<nets_date_2>\d{6})
            (?P<nets_date_3>\d{6})

            0{21}   # Filler
            $
            ''',
            re.VERBOSE,
        ),
        re.compile(
            r'''
            ^
            NY      # Format code
            (?P<service_code>21)
            (?P<assignment_type>24)     # AvtaleGiro agreements
            88      # Record type

            (?P<num_transactions>\d{8})
            (?P<num_records>\d{8})

            0{56}   # Filler
            $
            ''',
            re.VERBOSE,
        ),
        re.compile(
            r'''
            ^
            NY      # Format code
            (?P<service_code>21)
            (?P<assignment_type>36)     # AvtaleGiro cancellations
            88      # Record type

            (?P<num_transactions>\d{8})
            (?P<num_records>\d{8})
            (?P<total_amount>\d{17})
            (?P<nets_date_1>\d{6})
            (?P<nets_date_2>\d{6})

            0{27}   # Filler
            $
            ''',
            re.VERBOSE,
        ),
    ]

    @property
    def nets_date(self) -> Optional['datetime.date']:
        """Nets' processing date.

        Only used for OCR Giro.
        """
        return self.nets_date_1 if self.service_code == ServiceCode.OCR_GIRO else None

    @property
    def nets_date_earliest(self) -> Optional['datetime.date']:
        """Earliest date from the contained transactions."""
        if self.service_code == ServiceCode.OCR_GIRO:
            return self.nets_date_2
        elif self.service_code == ServiceCode.AVTALEGIRO:
            return self.nets_date_1
        else:  # pragma: no cover
            raise ValueError(f'Unhandled service code: {self.service_code}')

    @property
    def nets_date_latest(self) -> Optional['datetime.date']:
        """Latest date from the contained transactions."""
        if self.service_code == ServiceCode.OCR_GIRO:
            return self.nets_date_3
        elif self.service_code == ServiceCode.AVTALEGIRO:
            return self.nets_date_2
        else:  # pragma: no cover
            raise ValueError(f'Unhandled service code: {self.service_code}')

    def to_ocr(self) -> str:
        """Get record as OCR string."""
        return (
            'NY'
            f'{self.service_code:02d}'
            f'{self.assignment_type:02d}'
            '88'
            f'{self.num_transactions:08d}'
            f'{self.num_records:08d}'
            + (self.total_amount and f'{self.total_amount:017d}' or ('0' * 17))
            + (self.nets_date_1 and f'{self.nets_date_1:%d%m%y}' or ('0' * 6))
            + (self.nets_date_2 and f'{self.nets_date_2:%d%m%y}' or ('0' * 6))
            + (self.nets_date_3 and f'{self.nets_date_3:%d%m%y}' or ('0' * 6))
            + ('0' * 21)
        )


@define
class TransactionRecord(Record, ABC):
    """Transaction record base class."""

    transaction_type: 'TransactionType' = field(converter=to_transaction_type)
    transaction_number: int = field(converter=int)


@define
class TransactionAmountItem1(TransactionRecord):
    """TransactionAmountItem1 is the first record of a transaction.

    The record is used both for AvtaleGiro and for OCR Giro.
    """

    nets_date: 'datetime.date' = field(converter=to_date)
    amount: int = field(converter=int)
    kid: Optional[str] = field(
        converter=to_safe_str_or_none, validator=optional(str_of_max_length(25))
    )

    # Only OCR Giro
    centre_id: Optional[str] = field(default=None, validator=optional(str_of_length(2)))
    day_code: Optional[int] = field(default=None, converter=to_int_or_none)
    partial_settlement_number: Optional[int] = field(default=None, converter=to_int_or_none)
    partial_settlement_serial_number: Optional[str] = field(
        default=None, validator=optional(str_of_length(5))
    )
    sign: Optional[str] = field(default=None, validator=optional(str_of_length(1)))

    RECORD_TYPE: ClassVar[RecordType] = RecordType.TRANSACTION_AMOUNT_ITEM_1
    _PATTERNS: ClassVar[List[Pattern]] = [
        re.compile(
            r'''
            ^
            NY      # Format code
            (?P<service_code>09)
            (?P<transaction_type>\d{2})  # 10-21
            30      # Record type

            (?P<transaction_number>\d{7})
            (?P<nets_date>\d{6})

            (?P<centre_id>\d{2})
            (?P<day_code>\d{2})
            (?P<partial_settlement_number>\d{1})
            (?P<partial_settlement_serial_number>\d{5})
            (?P<sign>[-0]{1})

            (?P<amount>\d{17})
            (?P<kid>[\d ]{25})

            0{6}    # Filler
            $
            ''',
            re.VERBOSE,
        ),
        re.compile(
            r'''
            ^
            NY      # Format code
            (?P<service_code>21)
            (?P<transaction_type>\d{2})  # 02, 21, or 93
            30      # Record type

            (?P<transaction_number>\d{7})
            (?P<nets_date>\d{6})

            [ ]{11} # Filler

            (?P<amount>\d{17})
            (?P<kid>[\d ]{25})

            0{6}    # Filler
            $
            ''',
            re.VERBOSE,
        ),
    ]

    def to_ocr(self) -> str:
        """Get record as OCR string."""
        if self.service_code == ServiceCode.OCR_GIRO:
            ocr_giro_fields = (
                f'{self.centre_id:2}'
                f'{self.day_code:02d}'
                f'{self.partial_settlement_number:01d}'
                f'{self.partial_settlement_serial_number:5}'
                f'{self.sign:1}'
            )
        else:
            ocr_giro_fields = ' ' * 11

        return (
            'NY'
            f'{self.service_code:02d}'
            f'{self.transaction_type:02d}'
            '30'
            f'{self.transaction_number:07d}'
            f'{self.nets_date:%d%m%y}'
            + ocr_giro_fields
            + f'{self.amount:017d}{self.kid:>25}'
            + ('0' * 6)
        )


@define
class TransactionAmountItem2(TransactionRecord):
    """TransactionAmountItem2 is the second record of a transaction.

    The record is used both for AvtaleGiro and for OCR Giro.
    """

    # TODO Validate `reference` length, which depends on service code
    reference: Optional[str] = field(converter=to_safe_str_or_none)

    # Only OCR Giro
    form_number: Optional[str] = field(default=None, validator=optional(str_of_length(10)))
    bank_date: Optional['datetime.date'] = field(default=None, converter=to_date_or_none)
    debit_account: Optional[str] = field(default=None, validator=optional(str_of_length(11)))
    # XXX In use in OCR Giro "from giro debited account" transactions in test
    # data, but documented as a filler field.
    _filler: Optional[str] = field(default=None)

    # Only AvtaleGiro
    payer_name: Optional[str] = field(default=None, converter=to_safe_str_or_none)

    RECORD_TYPE: ClassVar[RecordType] = RecordType.TRANSACTION_AMOUNT_ITEM_2
    _PATTERNS: ClassVar[List[Pattern]] = [
        re.compile(
            r'''
            ^
            NY      # Format code
            (?P<service_code>09)
            (?P<transaction_type>\d{2})  # 10-21
            31      # Record type

            (?P<transaction_number>\d{7})
            (?P<form_number>\d{10})
            (?P<reference>\d{9})

            (?P<filler>.{7})  # XXX Documented as filler, in use in test data

            (?P<bank_date>\d{6})
            (?P<debit_account>\d{11})

            0{22}    # Filler
            $
            ''',
            re.VERBOSE,
        ),
        re.compile(
            r'''
            ^
            NY      # Format code
            (?P<service_code>21)
            (?P<transaction_type>\d{2})  # 02, 21, or 93
            31      # Record type

            (?P<transaction_number>\d{7})
            (?P<payer_name>.{10})

            [ ]{25} # Filler

            (?P<reference>.{25})

            0{5}    # Filler
            $
            ''',
            re.VERBOSE,
        ),
    ]

    def to_ocr(self) -> str:
        """Get record as OCR string."""
        common_fields = (
            f'NY{self.service_code:02d}{self.transaction_type:02d}31{self.transaction_number:07d}'
        )
        if self.service_code == ServiceCode.OCR_GIRO:
            service_fields = (
                f'{self.form_number:10}'
                + (self.reference and f'{self.reference:9}' or (' ' * 9))
                + (self._filler and f'{self._filler:7}' or ('0' * 7))
                + (self.bank_date and f'{self.bank_date:%d%m%y}' or '0' * 6)
                + f'{self.debit_account:11}'
                + ('0' * 22)
            )
        elif self.service_code == ServiceCode.AVTALEGIRO:
            service_fields = (
                (self.payer_name and f'{self.payer_name[:10]:10}' or (' ' * 10))
                + (' ' * 25)
                + (self.reference and f'{self.reference:25}' or (' ' * 25))
                + ('0' * 5)
            )
        else:  # pragma: no cover
            service_fields = ' ' * 35

        return common_fields + service_fields


@define
class TransactionAmountItem3(TransactionRecord):
    """TransactionAmountItem3 is the third record of a transaction.

    The record is only used for some OCR Giro transaction types.
    """

    text: Optional[str] = field(
        converter=to_safe_str_or_none, validator=optional(str_of_max_length(40))
    )

    RECORD_TYPE: ClassVar[RecordType] = RecordType.TRANSACTION_AMOUNT_ITEM_3
    _PATTERNS: ClassVar[List[Pattern]] = [
        re.compile(
            r'''
            ^
            NY      # Format code
            (?P<service_code>09)
            (?P<transaction_type>\d{2})  # 20-21
            32      # Record type

            (?P<transaction_number>\d{7})
            (?P<text>.{40})

            0{25}    # Filler
            $
            ''',
            re.VERBOSE,
        )
    ]

    def to_ocr(self) -> str:
        """Get record as OCR string."""
        return (
            f'NY09{self.transaction_type:02d}32{self.transaction_number:07d}'
            + (self.text and f'{self.text:40}' or (' ' * 40))
            + ('0' * 25)
        )


@define
class TransactionSpecification(TransactionRecord):
    """TransactionSpecification is used for AvtaleGiro transactions.

    The record is only used when bank notification is used to notify the payer.

    Each record contains half of an 80 char long line of text and can be
    repeated up to 84 times for a single transaction for a total of 42 lines of
    specification text.
    """

    line_number: int = field(converter=int)
    column_number: int = field(converter=int)
    text = field(
        converter=stripped_newlines(fixed_len_str(40, str)),  # type: ignore[misc]
        validator=optional(str_of_max_length(40)),
    )

    RECORD_TYPE: ClassVar[RecordType] = RecordType.TRANSACTION_SPECIFICATION
    _PATTERNS: ClassVar[List[Pattern]] = [
        re.compile(
            r'''
            ^
            NY      # Format code
            (?P<service_code>21)
            (?P<transaction_type>21)
            49      # Record type

            (?P<transaction_number>\d{7})
            4       # Payment notification
            (?P<line_number>\d{3})
            (?P<column_number>\d{1})
            (?P<text>.{40})

            0{20}    # Filler
            $
            ''',
            re.VERBOSE,
        )
    ]

    _MAX_LINES = 42
    _MAX_LINE_LENGTH = 80
    _MAX_COLUMNS = 2
    _MAX_RECORDS = _MAX_LINES * _MAX_COLUMNS

    @classmethod
    def from_text(
        cls,
        *,
        service_code: ServiceCode,
        transaction_type: 'TransactionType',
        transaction_number: int,
        text: Optional[str],
    ) -> Iterable['TransactionSpecification']:
        """Create a sequence of specification records from a text string."""
        for line, column, txt in cls._split_text_to_lines_and_columns(text or ''):
            yield cls(
                service_code=service_code,
                transaction_type=transaction_type,
                transaction_number=transaction_number,
                line_number=line,
                column_number=column,
                text=txt,
            )

    @classmethod
    def _split_text_to_lines_and_columns(cls, text: str) -> Iterable[Tuple[int, int, str]]:
        lines = text.splitlines()

        if len(lines) > cls._MAX_LINES:
            raise ValueError(f'Max {cls._MAX_LINES} specification lines allowed, got {len(lines)}')

        for line_number, line_text in enumerate(lines, 1):
            if len(line_text) > cls._MAX_LINE_LENGTH:
                raise ValueError(
                    'Specification lines must be max {} chars long, got {}: {!r}'.format(
                        cls._MAX_LINE_LENGTH, len(line_text), line_text
                    )
                )

            yield line_number, 1, f'{line_text[:40]:40}'
            yield line_number, 2, f'{line_text[40:80]:40}'

    @classmethod
    def to_text(cls, records: List['TransactionSpecification']) -> str:
        """Get a text string from a sequence of specification records."""
        if len(records) > cls._MAX_RECORDS:
            raise ValueError(
                f'Max {cls._MAX_RECORDS} specification records allowed, got {len(records)}'
            )

        tuples = sorted((r.line_number, r.column_number, r) for r in records)

        text = ''
        for _, column, specification in tuples:
            if specification.text:
                text += specification.text
                if column == cls._MAX_COLUMNS:
                    text += '\n'

        return text

    def to_ocr(self) -> str:
        """Get record as OCR string."""
        return (
            'NY212149'
            f'{self.transaction_number:07d}'
            '4'
            f'{self.line_number:03d}'
            f'{self.column_number:01d}'
            f'{self.text:40}'
            + ('0' * 20)
        )


@define
class AvtaleGiroAgreement(TransactionRecord):
    """AvtaleGiroAgreement is used by Nets to notify about agreement changes.

    This includes new or deleted agreements, as well as updates to the payer's
    notification preferences.
    """

    registration_type: 'AvtaleGiroRegistrationType' = field(
        converter=to_avtalegiro_registration_type
    )
    kid: Optional[str] = field(
        converter=to_safe_str_or_none, validator=optional(str_of_max_length(25))
    )
    notify: bool = field(converter=to_bool)

    RECORD_TYPE: ClassVar[RecordType] = RecordType.TRANSACTION_AGREEMENTS
    _PATTERNS: ClassVar[List[Pattern]] = [
        re.compile(
            r'''
            ^
            NY      # Format code
            (?P<service_code>21)
            (?P<transaction_type>94)
            70      # Record type

            (?P<transaction_number>\d{7})
            (?P<registration_type>\d{1})
            (?P<kid>[\d ]{25})
            (?P<notify>[JN]{1})

            0{38}   # Filler
            $
            ''',
            re.VERBOSE,
        )
    ]

    def to_ocr(self) -> str:
        """Get record as OCR string."""
        return (
            f'NY219470{self.transaction_number:07d}{self.registration_type:01d}{self.kid:>25}'
            + (self.notify and 'J' or 'N')
            + ('0' * 38)
        ).format(self=self)


def parse(data: str) -> List[R]:
    """Parse an OCR file into a list of record objects."""

    def all_subclasses(cls: Union[Type[R], Type[Record]]) -> List[Type[R]]:
        """Return a list of subclasses for a given class."""
        classes = cls.__subclasses__() + [
            subsubcls for subcls in cls.__subclasses__() for subsubcls in all_subclasses(subcls)
        ]
        return cast(List[Type[R]], classes)

    record_classes = {
        cls.RECORD_TYPE: cls for cls in all_subclasses(Record) if hasattr(cls, 'RECORD_TYPE')
    }

    results: List[R] = []

    for line in data.strip().splitlines():
        if len(line) != 80:
            raise ValueError('All lines must be exactly 80 chars long')

        record_type_str = line[6:8]
        if not record_type_str.isnumeric():
            raise ValueError(f'Record type must be numeric, got {record_type_str!r}')

        record_type = to_record_type(record_type_str)
        record_cls = record_classes[record_type]

        results.append(record_cls.from_string(line))

    return results



from contextlib import suppress
from datetime import timedelta
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from datetime import date, datetime

try:
    import zoneinfo
except ImportError:
    from backports import zoneinfo  # type: ignore[no-redef]

__all__ = ['get_minimum_due_date']

# Nets operates in the Norwegian timezone
OSLO_TZ = zoneinfo.ZoneInfo('Europe/Oslo')


def get_minimum_due_date(now: 'datetime') -> 'date':
    """
    Return the minimum valid due date for an avtalegiro ocrgiro created right now.

    The avtalegiro spec specifies that customers should have at least
    4 calendar days notice before a payment, so files containing due dates
    earlier than that will fail.

    "Calendar days" include weekends, but not holidays, so the holidays
    library is used to offset by holidays if it's installed.

    Logic is used for validation in netsgiro internally, but is also exported
    to be used downstream, when generating OCR files.
    """
    today = now.date()

    # 14:00 is the cut-off for sending in new transmissions;
    # files sent after 14:00 are processed the next day.
    delta = 4 if now.hour < 14 else 5

    # Adjust for holidays, if the dependency is installed
    # - Users of the library that want this, should install
    # - netsgiro with `pip install netsgiro[holidays]
    with suppress(ImportError):
        from holidays import country_holidays

        # Get holidays
        holidays_in_the_date_range = country_holidays('NO')[now : now + timedelta(days=delta)]  # type: ignore[misc]

        # Add days to `delta` for each holiday found in the date range
        delta += len(holidays_in_the_date_range)

    # Calendar days don't count weekends, but file do have
    # to be received on weekdays to be processed the same day
    if today.weekday() in [5, 6]:
        delta += 7 - today.weekday()

    return (now + timedelta(days=delta)).date()



"""Custom validators for :mod:`attrs`."""
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any, Callable

from attr import Attribute
from attr.validators import instance_of

from netsgiro.utils import OSLO_TZ, get_minimum_due_date

if TYPE_CHECKING:
    from datetime import date


C = Callable[[object, Attribute, Any], None]


def str_of_length(length: int) -> C:
    """Validate that the value is a string of the given length."""

    def validator(instance: object, attribute: Attribute, value: Any) -> None:
        instance_of(str)(instance, attribute, value)
        if len(value) != length:
            raise ValueError(f'{attribute.name} must be exactly {length} chars, got {value!r}')

    return validator


def str_of_max_length(length: int) -> C:
    """Validate that the value is a string with a max length."""

    def validator(instance: object, attribute: Attribute, value: Any) -> None:
        instance_of(str)(instance, attribute, value)
        if len(value) > length:
            raise ValueError(
                '{0.name} must be at most {1} chars, got {2!r} which is {3} chars'.format(
                    attribute, length, value, len(value)
                )
            )

    return validator


def validate_due_date(value: 'date') -> None:
    """Make sure payment request dates are gt the minimum allowed date."""
    now = datetime.now(tz=OSLO_TZ)

    if value < get_minimum_due_date(now=now):
        raise ValueError(
            'The minimum due date of a transaction is today + 4 calendar days.'
            ' OCR files with due dates earlier than this will be rejected when'
            ' submitted.'
        )

    if value > (now + timedelta(days=365)).date():
        raise ValueError(
            'The maximum due date of a transaction is 12 months in the future.'
            ' OCR files with due dates later than this will be rejected when'
            ' submitted.'
        )



"""File parsers for Nets AvtaleGiro and OCR Giro files."""

from typing import List

from netsgiro.constants import *
from netsgiro.enums import *
from netsgiro.objects import *
from netsgiro.utils import *

from netsgiro import constants, enums, objects, utils  # isort: skip

__version__ = '1.3.0'

__all__: List[str] = constants.__all__ + enums.__all__ + objects.__all__ + utils.__all__



"""Enums for all codes used in OCR files."""

from enum import IntEnum
from typing import List

__all__: List[str] = [
    'ServiceCode',
    'RecordType',
    'AssignmentType',
    'TransactionType',
    'AvtaleGiroRegistrationType',
]


class ServiceCode(IntEnum):
    """Service codes tell which Nets service the record applies to."""

    #: Used for the transmission start and end record.
    NONE = 0

    #: Used for all OCR Giro records.
    OCR_GIRO = 9

    #: Used for all AvtaleGiro records.
    AVTALEGIRO = 21


class RecordType(IntEnum):
    """Record types tell what type of record this is."""

    #: See :class:`netsgiro.TransmissionStart`.
    TRANSMISSION_START = 10

    #: See :class:`netsgiro.AssignmentStart`.
    ASSIGNMENT_START = 20

    #: See :class:`netsgiro.TransactionAmountItem1`.
    TRANSACTION_AMOUNT_ITEM_1 = 30

    #: See :class:`netsgiro.TransactionAmountItem2`.
    TRANSACTION_AMOUNT_ITEM_2 = 31

    #: See :class:`netsgiro.TransactionAmountItem3`.
    TRANSACTION_AMOUNT_ITEM_3 = 32

    #: See :class:`netsgiro.TransactionSpecification`.
    TRANSACTION_SPECIFICATION = 49

    #: See :class:`netsgiro.TransactionAgreements`.
    TRANSACTION_AGREEMENTS = 70

    #: See :class:`netsgiro.AssignmentEnd`.
    ASSIGNMENT_END = 88

    #: See :class:`netsgiro.TransmissionEnd`.
    TRANSMISSION_END = 89


class AssignmentType(IntEnum):
    """Assignment types tell what type of assignment this is."""

    #: Used both for AvtaleGiro payment requests and OCR Giro transactions.
    TRANSACTIONS = 0

    #: Used for AvtaleGiro agreement updates.
    AVTALEGIRO_AGREEMENTS = 24

    #: Used for AvtaleGiro cancellations.
    AVTALEGIRO_CANCELLATIONS = 36


class TransactionType(IntEnum):
    """Assignment types tell what type of transaction this is."""

    #: Used for OCR Giro.
    FROM_GIRO_DEBITED_ACCOUNT = 10
    #: Used for OCR Giro.
    FROM_STANDING_ORDERS = 11
    #: Used for OCR Giro.
    FROM_DIRECT_REMITTANCE = 12
    #: Used for OCR Giro.
    FROM_BUSINESS_TERMINAL_GIRO = 13
    #: Used for OCR Giro.
    FROM_COUNTER_GIRO = 14
    #: Used for OCR Giro.
    FROM_AVTALEGIRO = 15
    #: Used for OCR Giro.
    FROM_TELEGIRO = 16
    #: Used for OCR Giro.
    FROM_CASH_GIRO = 17

    #: Used for OCR Giro.
    REVERSING_WITH_KID = 18
    #: Used for OCR Giro.
    PURCHASE_WITH_KID = 19
    #: Used for OCR Giro.
    REVERSING_WITH_TEXT = 20
    #: Used for OCR Giro.
    #:
    #: .. note::
    #:     The value ``21`` is used for both :attr:`PURCHASE_WITH_TEXT` and
    #:     :attr:`AVTALEGIRO_WITH_BANK_NOTIFICATION`. The enum
    #:     representation will be ``<PURCHASE_WITH_TEXT: 21>`` in either case.
    PURCHASE_WITH_TEXT = 21

    #: Used for AvtaleGiro when you want to notify the payer yourself.
    AVTALEGIRO_WITH_PAYEE_NOTIFICATION = 2
    #: Used for AvtaleGiro when you want the bank to notify the payer.
    AVTALEGIRO_WITH_BANK_NOTIFICATION = 21
    #: Used for transactions that are part of an AvtaleGiro cancellation
    #: assignment.
    AVTALEGIRO_CANCELLATION = 93
    #: Used by Nets for updates to AvtaleGiro agreeements.
    AVTALEGIRO_AGREEMENT = 94


class AvtaleGiroRegistrationType(IntEnum):
    """AvtaleGiro registration types tell what kind of update this is."""

    #: Used when the AvtaleGiro agreement assignment contains all currently
    #: active agreements.
    ACTIVE_AGREEMENT = 0

    #: Used when the AvtaleGiro agreement assignment contains only changes,
    #: and the current agreement is new or updated.
    NEW_OR_UPDATED_AGREEMENT = 1

    #: Used when the AvtaleGiro agreement assignment contains only changes,
    #: and the current agreement has been deleted.
    DELETED_AGREEMENT = 2



"""The higher-level objects API."""

import datetime
from collections import OrderedDict
from decimal import Decimal
from typing import TYPE_CHECKING, Callable, ClassVar, Iterable, List, Optional
from typing import OrderedDict as OrderedDictType
from typing import TypeVar, Union, cast

from attrs import Factory, define, field
from attrs.validators import instance_of, optional

from netsgiro.converters import (
    to_assignment_type,
    to_avtalegiro_registration_type,
    to_service_code,
    to_transaction_type,
)
from netsgiro.enums import AssignmentType, ServiceCode, TransactionType
from netsgiro.records import (
    AssignmentEnd,
    AssignmentStart,
    AvtaleGiroAgreement,
    TransactionAmountItem1,
    TransactionAmountItem2,
    TransactionAmountItem3,
    TransactionRecord,
    TransactionSpecification,
    TransmissionEnd,
    TransmissionStart,
)
from netsgiro.records import parse as records_parse
from netsgiro.validators import str_of_length
from netsgiro.validators import validate_due_date as _validate_due_date

if TYPE_CHECKING:
    from netsgiro.enums import AvtaleGiroRegistrationType
    from netsgiro.records import Record

__all__: List[str] = [
    'Transmission',
    'Assignment',
    'Agreement',
    'PaymentRequest',
    'Transaction',
    'parse',
]

TransactionAmountItems = Union[
    TransactionAmountItem1, TransactionAmountItem2, TransactionAmountItem3
]

# Record or Record subclasses
R = TypeVar('R', bound='Record')


@define
class Transmission:
    """Transmission is the top-level object.

    An OCR file contains a single transmission. The transmission can contain
    multiple :class:`~netsgiro.Assignment` objects of various types.
    """

    #: Data transmitters unique enumeration of the transmission. String of 7 digits.
    number: str = field(validator=str_of_length(7))

    #: Data transmitter's Nets ID. String of 8 digits.
    data_transmitter: str = field(validator=str_of_length(8))

    #: Data recipient's Nets ID. String of 8 digits.
    data_recipient: str = field(validator=str_of_length(8))

    #: For OCR Giro files from Nets, this is Nets' processing date.
    #:
    #: For AvtaleGiro payment request, the earliest due date in the
    #: transmission is automatically used.
    date: Optional[datetime.date] = field(
        default=None, validator=optional(instance_of(datetime.date))
    )

    #: List of assignments.
    assignments: List['Assignment'] = field(default=Factory(list), repr=False)

    @classmethod
    def from_records(cls, records: List[R]) -> 'Transmission':
        """Build a Transmission object from a list of record objects."""
        if len(records) < 2:
            raise ValueError(f'At least 2 records required, got {len(records)}')

        start, body, end = records[0], records[1:-1], records[-1]

        assert isinstance(start, TransmissionStart)
        assert isinstance(end, TransmissionEnd)

        return cls(
            number=start.transmission_number,
            data_transmitter=start.data_transmitter,
            data_recipient=start.data_recipient,
            date=end.nets_date,
            assignments=cls._get_assignments(body),
        )

    @staticmethod
    def _get_assignments(records: List[R]) -> List['Assignment']:
        assignments: OrderedDictType[str, List[R]] = OrderedDict()

        current_assignment_number = None
        for record in records:
            if isinstance(record, AssignmentStart):
                current_assignment_number = record.assignment_number
                assignments[current_assignment_number] = []
            if current_assignment_number is None:
                raise ValueError(f'Expected AssignmentStart record, got {record!r}')
            assignments[current_assignment_number].append(record)
            if isinstance(record, AssignmentEnd):
                current_assignment_number = None

        return [Assignment.from_records(rs) for rs in assignments.values()]

    def to_ocr(self) -> str:
        """Convert the transmission to an OCR string."""
        lines = [record.to_ocr() for record in self.to_records()]
        return '\n'.join(lines)

    def to_records(self) -> Iterable['Record']:
        """Convert the transmission to a list of records."""
        yield self._get_start_record()
        for assignment in self.assignments:
            yield from assignment.to_records()
        yield self._get_end_record()

    def _get_start_record(self) -> 'Record':
        return TransmissionStart(
            service_code=ServiceCode.NONE,
            transmission_number=self.number,
            data_transmitter=self.data_transmitter,
            data_recipient=self.data_recipient,
        )

    def _get_end_record(self) -> 'Record':
        avtalegiro_payment_request = all(
            assignment.service_code == ServiceCode.AVTALEGIRO
            and assignment.type
            in (
                AssignmentType.TRANSACTIONS,
                AssignmentType.AVTALEGIRO_CANCELLATIONS,
            )
            for assignment in self.assignments
        )
        if self.assignments and avtalegiro_payment_request:
            date: Optional[datetime.date] = None
            for assignment in self.assignments:
                earliest_transaction_date = assignment.get_earliest_transaction_date()
                if date is None or (
                    isinstance(earliest_transaction_date, datetime.date)
                    and earliest_transaction_date < date
                ):
                    date = earliest_transaction_date
        else:
            date = self.date

        assert isinstance(date, datetime.date)

        return TransmissionEnd(
            service_code=ServiceCode.NONE,
            num_transactions=self.get_num_transactions(),
            num_records=self.get_num_records(),
            total_amount=int(self.get_total_amount() * 100),
            nets_date=date,
        )

    def add_assignment(
        self,
        *,
        service_code: ServiceCode,
        assignment_type: AssignmentType,
        agreement_id: Optional[str] = None,
        number: str,
        account: str,
        date: Optional[datetime.date] = None,
    ) -> 'Assignment':
        """Add an assignment to the transmission."""
        assignment = Assignment(
            service_code=service_code,
            type=assignment_type,
            agreement_id=agreement_id,
            number=number,
            account=account,
            date=date,
        )
        self.assignments.append(assignment)
        return assignment

    def get_num_transactions(self) -> int:
        """Get number of transactions in the transmission."""
        return sum(assignment.get_num_transactions() for assignment in self.assignments)

    def get_num_records(self) -> int:
        """Get number of records in the transmission.

        Includes the transmission's start and end record.
        """
        return 2 + sum(assignment.get_num_records() for assignment in self.assignments)

    def get_total_amount(self) -> Decimal:
        """Get the total amount from all transactions in the transmission."""
        return Decimal(sum(assignment.get_total_amount() for assignment in self.assignments))


# Assigment transactions
TS = List[Union['Transaction', 'Agreement', 'PaymentRequest']]

# TransactionRecord and subclasses
TR = TypeVar('TR', bound=TransactionRecord)


@define
class Assignment:
    """An Assignment groups multiple transactions within a transmission.

    Use :meth:`netsgiro.Transmission.add_assignment` to create assignments.
    """

    #: The service code. One of :class:`~netsgiro.ServiceCode`.
    service_code: ServiceCode = field(converter=to_service_code)

    #: The transaction type. One of :class:`~TransactionType`.
    type: AssignmentType = field(converter=to_assignment_type)

    #: The assignment number. String of 7 digits.
    number: str = field(validator=str_of_length(7))

    #: The payee's bank account. String of 11 digits.
    account: str = field(validator=str_of_length(11))

    #: Used for OCR Giro.
    #:
    #: The payee's agreement ID with Nets. String of 9 digits.
    agreement_id: Optional[str] = field(default=None, validator=optional(str_of_length(9)))

    #: Used for OCR Giro.
    #:
    #: The date the assignment was generated by Nets.
    date: Optional[datetime.date] = field(
        default=None, validator=optional(instance_of(datetime.date))
    )

    #: List of transaction objects, like :class:`~netsgiro.Agreement`,
    #: :class:`~netsgiro.PaymentRequest`, :class:`~netsgiro.Transaction`.
    transactions: TS = field(default=Factory(list), repr=False)

    _next_transaction_number: int = field(default=1, init=False)

    @classmethod
    def from_records(cls, records: List[R]) -> 'Assignment':
        """Build an Assignment object from a list of record objects."""
        if len(records) < 2:
            raise ValueError(f'At least 2 records required, got {len(records)}')

        start, body, end = records[0], records[1:-1], records[-1]

        assert isinstance(start, AssignmentStart)
        assert isinstance(end, AssignmentEnd)

        sc = start.service_code
        at = start.assignment_type

        cb: Callable
        if sc == ServiceCode.OCR_GIRO:
            # -> List[Transaction]
            cb = cls._get_transactions
        elif sc == ServiceCode.AVTALEGIRO:
            if at == AssignmentType.AVTALEGIRO_AGREEMENTS:
                # -> List[Agreement]
                cb = cls._get_agreements
            else:
                # -> List[PaymentRequest]
                cb = cls._get_payment_requests
        else:  # pragma: no cover
            raise ValueError(f'Unknown service code: {start.service_code}')

        transactions: TS = cb(body)

        return cls(
            service_code=start.service_code,
            type=start.assignment_type,
            agreement_id=start.agreement_id,
            number=start.assignment_number,
            account=start.assignment_account,
            date=end.nets_date,
            transactions=transactions,
        )

    @staticmethod
    def _get_agreements(records: List[AvtaleGiroAgreement]) -> List['Agreement']:
        return [Agreement.from_records([r]) for r in records]

    @classmethod
    def _get_payment_requests(cls, records: List[TR]) -> List['PaymentRequest']:
        transactions = cls._group_by_transaction_number(records)
        return [PaymentRequest.from_records(rs) for rs in transactions.values()]

    @classmethod
    def _get_transactions(cls, records: List[TR]) -> List['Transaction']:
        transactions = cls._group_by_transaction_number(records)
        return [Transaction.from_records(rs) for rs in transactions.values()]

    @staticmethod
    def _group_by_transaction_number(records: List[TR]) -> OrderedDictType[int, List[TR]]:
        transactions: OrderedDictType[int, List[TR]] = OrderedDict()
        for transaction_record in records:
            tr = transaction_record.transaction_number
            if tr not in transactions:
                transactions[tr] = []
            transactions[tr].append(transaction_record)

        return transactions

    def to_records(self) -> Iterable[Union[AssignmentStart, TransactionAmountItems, AssignmentEnd]]:
        """Convert the assignment to a list of records."""
        yield self._get_start_record()
        for transaction in self.transactions:
            yield from transaction.to_records()
        yield self._get_end_record()

    def _get_start_record(self) -> AssignmentStart:
        return AssignmentStart(
            service_code=self.service_code,
            assignment_type=self.type,
            assignment_number=self.number,
            assignment_account=self.account,
            agreement_id=self.agreement_id,
        )

    def _get_end_record(self) -> AssignmentEnd:
        if self.service_code == ServiceCode.OCR_GIRO:
            dates = {
                'nets_date_1': self.date,
                'nets_date_2': self.get_earliest_transaction_date(),
                'nets_date_3': self.get_latest_transaction_date(),
            }
        elif self.service_code == ServiceCode.AVTALEGIRO:
            dates = {
                'nets_date_1': self.get_earliest_transaction_date(),
                'nets_date_2': self.get_latest_transaction_date(),
            }
        else:  # pragma: no cover
            raise ValueError(f'Unhandled service code: {self.service_code}')

        return AssignmentEnd(
            service_code=self.service_code,
            assignment_type=self.type,
            num_transactions=self.get_num_transactions(),
            num_records=self.get_num_records(),
            total_amount=int(self.get_total_amount() * 100),
            **dates,
        )

    def add_payment_request(
        self,
        *,
        kid: str,
        due_date: datetime.date,
        amount: Decimal,
        reference: Optional[str] = None,
        payer_name: Optional[str] = None,
        bank_notification: Union[bool, str] = False,
        validate_due_date: bool = False,
    ) -> 'PaymentRequest':
        """Add an AvtaleGiro payment request to the assignment.

        The assignment must have service code
        :attr:`~ServiceCode.AVTALEGIRO` and assignment type
        :attr:`~AssignmentType.TRANSACTIONS`.
        """
        assert (
            self.service_code == ServiceCode.AVTALEGIRO
        ), 'Can only add payment requests to AvtaleGiro assignments'
        assert (
            self.type == AssignmentType.TRANSACTIONS
        ), 'Can only add payment requests to transaction assignments'

        if validate_due_date:
            # Make sure we're not passing invalid due dates
            _validate_due_date(due_date)

        if bank_notification:
            transaction_type = TransactionType.AVTALEGIRO_WITH_BANK_NOTIFICATION
        else:
            transaction_type = TransactionType.AVTALEGIRO_WITH_PAYEE_NOTIFICATION

        return self._add_avtalegiro_transaction(
            transaction_type=transaction_type,
            kid=kid,
            due_date=due_date,
            amount=amount,
            reference=reference,
            payer_name=payer_name,
            bank_notification=bank_notification,
        )

    def add_payment_cancellation(
        self,
        *,
        kid: str,
        due_date: datetime.date,
        amount: Decimal,
        reference: Optional[str] = None,
        payer_name: Optional[str] = None,
        bank_notification: Union[bool, str] = False,
    ) -> 'PaymentRequest':
        """Add an AvtaleGiro cancellation to the assignment.

        The assignment must have service code
        :attr:`~netsgiro.ServiceCode.AVTALEGIRO` and assignment type
        :attr:`~AssignmentType.AVTALEGIRO_CANCELLATIONS`.

        Otherwise, the cancellation must be identical to the payment request it
        is cancelling.
        """
        assert (
            self.service_code == ServiceCode.AVTALEGIRO
        ), 'Can only add cancellation to AvtaleGiro assignments'
        assert (
            self.type == AssignmentType.AVTALEGIRO_CANCELLATIONS
        ), 'Can only add cancellation to cancellation assignments'

        return self._add_avtalegiro_transaction(
            transaction_type=TransactionType.AVTALEGIRO_CANCELLATION,
            kid=kid,
            due_date=due_date,
            amount=amount,
            reference=reference,
            payer_name=payer_name,
            bank_notification=bank_notification,
        )

    def _add_avtalegiro_transaction(
        self,
        *,
        transaction_type: TransactionType,
        kid: str,
        due_date: datetime.date,
        amount: Decimal,
        reference: Optional[str] = None,
        payer_name: Optional[str] = None,
        bank_notification: Union[str, bool] = False,
    ) -> 'PaymentRequest':

        text = bank_notification if isinstance(bank_notification, str) else ''
        number = self._next_transaction_number
        self._next_transaction_number += 1

        transaction = PaymentRequest(
            service_code=self.service_code,
            type=transaction_type,
            number=number,
            date=due_date,
            amount=amount,
            kid=kid,
            reference=reference,
            text=text,
            payer_name=payer_name,
        )
        self.transactions.append(transaction)
        return transaction

    def get_num_transactions(self) -> int:
        """Get number of transactions in the assignment."""
        return len(self.transactions)

    def get_num_records(self) -> int:
        """Get number of records in the assignment.

        Includes the assignment's start and end record.
        """
        return 2 + sum(len(list(transaction.to_records())) for transaction in self.transactions)

    def get_total_amount(self) -> Decimal:
        """Get the total amount from all transactions in the assignment."""
        total = Decimal(0)
        for t in self.transactions:
            iter_amount = getattr(t, 'amount', None)
            if iter_amount:
                total += iter_amount
        return total

    def _get_earliest_or_latest_transaction_date(self, latest: bool) -> Optional[datetime.date]:
        """Get earliest or latest date from the assignment's transactions."""
        date_: Optional[datetime.date] = None
        for t in self.transactions:
            iter_date = getattr(t, 'date', None)
            if iter_date and (
                date_ is None or (iter_date > date_ if latest else iter_date < date_)
            ):
                date_ = iter_date
        return date_

    def get_earliest_transaction_date(self) -> Optional[datetime.date]:
        """Get earliest date from the assignment's transactions."""
        return self._get_earliest_or_latest_transaction_date(latest=False)

    def get_latest_transaction_date(self) -> Optional[datetime.date]:
        """Get latest date from the assignment's transactions."""
        return self._get_earliest_or_latest_transaction_date(latest=True)


@define
class Agreement:
    """Agreement contains an AvtaleGiro agreement update.

    Agreements are only found in assignments of the
    :attr:`~AssignmentType.AVTALEGIRO_AGREEMENTS` type, which are only
    created by Nets.
    """

    #: The service code. One of :class:`~netsgiro.ServiceCode`.
    service_code: ServiceCode = field(converter=to_service_code)

    #: Transaction number. Unique and ordered within an assignment.
    number: int = field(validator=instance_of(int))

    #: Type of agreement registration update.
    #: One of :class:`~AvtaleGiroRegistrationType`.
    registration_type: 'AvtaleGiroRegistrationType' = field(
        converter=to_avtalegiro_registration_type
    )

    #: KID number to identify the customer and invoice.
    kid: Optional[str] = field(validator=optional(instance_of(str)))

    #: Whether the payer wants notification about payment requests.
    notify: bool = field(validator=instance_of(bool))

    TRANSACTION_TYPE: ClassVar[TransactionType] = TransactionType.AVTALEGIRO_AGREEMENT

    @classmethod
    def from_records(cls, records: List[AvtaleGiroAgreement]) -> 'Agreement':
        """Build an Agreement object from a list of record objects."""
        assert len(records) == 1
        record = records[0]
        assert isinstance(record, AvtaleGiroAgreement)
        assert record.transaction_type == TransactionType.AVTALEGIRO_AGREEMENT

        return cls(
            service_code=record.service_code,
            number=record.transaction_number,
            registration_type=record.registration_type,
            kid=record.kid,
            notify=record.notify,
        )

    def to_records(self) -> Iterable[AvtaleGiroAgreement]:
        """Convert the agreement to a list of records."""
        yield AvtaleGiroAgreement(
            service_code=self.service_code,
            transaction_type=self.TRANSACTION_TYPE,
            transaction_number=self.number,
            registration_type=self.registration_type,
            kid=self.kid,
            notify=self.notify,
        )


@define
class PaymentRequest:
    """PaymentRequest contains an AvtaleGiro payment request or cancellation.

    To create a transaction, you will normally use the helper methods on
    :class:`~netsgiro.Assignment`:
    :meth:`~netsgiro.Assignment.add_payment_request` and
    :meth:`~netsgiro.Assignment.add_payment_cancellation`.
    """

    #: The service code. One of :class:`~netsgiro.ServiceCode`.
    service_code: ServiceCode = field(converter=to_service_code)

    #: The transaction type. One of :class:`~TransactionType`.
    type: TransactionType = field(converter=to_transaction_type)

    #: Transaction number. Unique and ordered within an assignment.
    number: int = field(validator=instance_of(int))

    #: The due date.
    date: datetime.date = field(validator=instance_of(datetime.date))

    #: Transaction amount in NOK with two decimals.
    amount: Decimal = field(converter=Decimal)

    #: KID number to identify the customer and invoice.
    kid: Optional[str] = field(validator=optional(instance_of(str)))

    #: This is a specification line that will, if set, be displayed on the
    #: payers account statement. Alphanumeric, max 25 chars.
    reference: Optional[str] = field(validator=optional(instance_of(str)))

    #: This is up to 42 lines of 80 chars each of free text used by the bank to
    #: notify the payer about the payment request. It is not used if the payee
    #: is responsible for notifying the payer.
    text: Optional[str] = field(validator=optional(instance_of(str)))

    #: The value is only used to help the payee cross-reference reports from
    #: Nets with their own records. It is not vi.attr.ible to the payer.
    payer_name: Optional[str] = field(validator=optional(instance_of(str)))

    @property
    def amount_in_cents(self) -> int:
        """Transaction amount in NOK cents."""
        return int(self.amount * 100)

    @classmethod
    def from_records(cls, records: List[TR]) -> 'PaymentRequest':
        """Build a Transaction object from a list of record objects."""
        amount_item_1 = records.pop(0)
        assert isinstance(amount_item_1, TransactionAmountItem1)
        amount_item_2 = records.pop(0)
        assert isinstance(amount_item_2, TransactionAmountItem2)

        text = TransactionSpecification.to_text(cast(List['TransactionSpecification'], records))

        return cls(
            service_code=amount_item_1.service_code,
            type=amount_item_1.transaction_type,
            number=amount_item_1.transaction_number,
            date=amount_item_1.nets_date,
            amount=Decimal(amount_item_1.amount) / 100,
            kid=amount_item_1.kid,
            reference=amount_item_2.reference,
            text=text,
            payer_name=amount_item_2.payer_name,
        )

    def to_records(
        self,
    ) -> Iterable[Union[TransactionAmountItem1, TransactionAmountItem2, TransactionSpecification]]:
        """Convert the transaction to a list of records."""
        yield TransactionAmountItem1(
            service_code=self.service_code,
            transaction_type=self.type,
            transaction_number=self.number,
            nets_date=self.date,
            amount=self.amount_in_cents,
            kid=self.kid,
        )
        yield TransactionAmountItem2(
            service_code=self.service_code,
            transaction_type=self.type,
            transaction_number=self.number,
            reference=self.reference,
            payer_name=self.payer_name,
        )

        if self.type == TransactionType.AVTALEGIRO_WITH_BANK_NOTIFICATION:
            yield from TransactionSpecification.from_text(
                service_code=self.service_code,
                transaction_type=self.type,
                transaction_number=self.number,
                text=self.text,
            )


@define
class Transaction:
    """Transaction contains an OCR Giro transaction.

    Transactions are found in assignments with the service code
    :attr:`~netsgiro.ServiceCode.OCR_GIRO` type, which are only
    created by Nets.
    """

    #: The service code. One of :class:`~netsgiro.ServiceCode`.
    service_code: ServiceCode = field(converter=to_service_code)

    #: The transaction type. One of :class:`~TransactionType`.
    type: TransactionType = field(converter=to_transaction_type)

    #: Transaction number. Unique and ordered within an assignment.
    number: int = field(validator=instance_of(int))

    #: Nets' processing date.
    date: datetime.date = field(validator=instance_of(datetime.date))

    #: Transaction amount in NOK with two decimals.
    amount: Decimal = field(converter=Decimal)

    #: KID number to identify the customer and invoice.
    kid: Optional[str] = field(validator=optional(instance_of(str)))

    #: The value depends on the payment method.
    reference: Optional[str] = field(validator=optional(instance_of(str)))

    #: Up to 40 chars of free text from the payment terminal.
    text: Optional[str] = field(validator=optional(instance_of(str)))

    #: Used for OCR Giro.
    centre_id: Optional[str] = field(validator=optional(str_of_length(2)))

    #: Used for OCR Giro.
    day_code: Optional[int] = field(validator=optional(instance_of(int)))

    #: Used for OCR Giro.
    partial_settlement_number: Optional[int] = field(validator=optional(instance_of(int)))

    #: Used for OCR Giro.
    partial_settlement_serial_number: Optional[str] = field(validator=optional(str_of_length(5)))

    #: Used for OCR Giro.
    sign: Optional[str] = field(validator=optional(str_of_length(1)))

    #: Used for OCR Giro.
    form_number: Optional[str] = field(validator=optional(str_of_length(10)))

    #: Used for OCR Giro.
    bank_date: Optional[datetime.date] = field(validator=optional(instance_of(datetime.date)))

    #: Used for OCR Giro.
    debit_account: Optional[str] = field(validator=optional(str_of_length(11)))

    _filler: Optional[str] = field(validator=optional(str_of_length(7)))

    @property
    def amount_in_cents(self) -> int:
        """Transaction amount in NOK cents."""
        return int(self.amount * 100)

    @classmethod
    def from_records(cls, records: List[TR]) -> 'Transaction':
        """Build a Transaction object from a list of record objects."""
        amount_item_1 = records.pop(0)
        assert isinstance(amount_item_1, TransactionAmountItem1)
        amount_item_2 = records.pop(0)
        assert isinstance(amount_item_2, TransactionAmountItem2)

        if len(records) == 1 and isinstance(records[0], TransactionAmountItem3):
            text = records[0].text
        else:
            text = None

        return cls(
            service_code=amount_item_1.service_code,
            type=amount_item_1.transaction_type,
            number=amount_item_1.transaction_number,
            date=amount_item_1.nets_date,
            amount=Decimal(amount_item_1.amount) / 100,
            kid=amount_item_1.kid,
            reference=amount_item_2.reference,
            text=text,
            centre_id=amount_item_1.centre_id,
            day_code=amount_item_1.day_code,
            partial_settlement_number=amount_item_1.partial_settlement_number,
            partial_settlement_serial_number=amount_item_1.partial_settlement_serial_number,
            sign=amount_item_1.sign,
            form_number=amount_item_2.form_number,
            bank_date=amount_item_2.bank_date,
            debit_account=amount_item_2.debit_account,
            filler=amount_item_2._filler,
        )

    def to_records(self) -> Iterable[TransactionAmountItems]:
        """Convert the transaction to a list of records."""
        yield TransactionAmountItem1(
            service_code=self.service_code,
            transaction_type=self.type,
            transaction_number=self.number,
            nets_date=self.date,
            amount=self.amount_in_cents,
            kid=self.kid,
            centre_id=self.centre_id,
            day_code=self.day_code,
            partial_settlement_number=self.partial_settlement_number,
            partial_settlement_serial_number=self.partial_settlement_serial_number,
            sign=self.sign,
        )
        yield TransactionAmountItem2(
            service_code=self.service_code,
            transaction_type=self.type,
            transaction_number=self.number,
            reference=self.reference,
            form_number=self.form_number,
            bank_date=self.bank_date,
            debit_account=self.debit_account,
            filler=self._filler,
        )

        if self.type in (
            TransactionType.REVERSING_WITH_TEXT,
            TransactionType.PURCHASE_WITH_TEXT,
        ):
            yield TransactionAmountItem3(
                service_code=self.service_code,
                transaction_type=self.type,
                transaction_number=self.number,
                text=self.text,
            )


def parse(data: str) -> Transmission:
    """Parse an OCR file into a Transmission object."""
    return Transmission.from_records(records_parse(data))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['netsgiro']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=17.4']

extras_require = \
{':python_version < "3.9"': ['backports.zoneinfo>=0.2.1,<0.3.0']}

setup_kwargs = {
    'name': 'netsgiro',
    'version': '2.0.0',
    'description': 'File parsers for Nets AvtaleGiro and OCR Giro',
    'long_description': '.. image:: https://img.shields.io/pypi/v/netsgiro.svg?style=flat\n    :target: https://pypi.org/project/netsgiro/\n    :alt: Latest PyPI version\n\n.. image:: https://github.com/otovo/python-netsgiro/actions/workflows/test.yml/badge.svg\n    :target: https://github.com/otovo/python-netsgiro/actions/workflows/test.yml\n    :alt: Github actions test pipeline status\n\n.. image:: https://img.shields.io/readthedocs/netsgiro.svg\n   :target: https://netsgiro.readthedocs.io/\n   :alt: Read the Docs build status\n\n.. image:: https://img.shields.io/codecov/c/github/otovo/python-netsgiro/master.svg\n   :target: https://codecov.io/gh/otovo/python-netsgiro\n   :alt: Test coverage\n\n========\nnetsgiro\n========\n\nnetsgiro is a Python library for working with `Nets <https://www.nets.eu/>`_\nAvtaleGiro and OCR Giro files.\n\nAvtaleGiro is a direct debit solution that is in widespread use in Norway, with\nmore than 15 000 companies offering it to their customers. OCR Giro is used by\nNets and Norwegian banks to update payees on recent deposits to their bank\naccounts. In combination, AvtaleGiro and OCR Giro allows for a high level of\nautomation of invoicing and payment processing.\n\nThe netsgiro library supports:\n\n- Parsing AvtaleGiro agreements\n- Creating AvtaleGiro payment requests\n- Creating AvtaleGiro cancellations\n- Parsing OCR Giro transactions\n\nnetsgiro is available from PyPI. To install it, run::\n\n    pip install netsgiro\n\nFor details and code examples, see `the netsgiro documentation\n<https://netsgiro.readthedocs.io/>`_.\n\nFor further details, please refer to the official\n`AvtaleGiro <https://www.avtalegiro.no/>`_ and\n`OCR Giro <https://www.nets.eu/no-nb/losninger/inn-og-utbetalinger/ocrgiro/Pages/default.aspx>`_\ndocumentation from Nets.\n\n\nLicense\n=======\n\nCopyright 2017-2019 `Otovo AS <https://www.otovo.com/>`_. Licensed under the\nApache License, Version 2.0. See the ``LICENSE`` file for the full license\ntext.\n\n\nProject resources\n=================\n\n- `Documentation <https://netsgiro.readthedocs.io/>`_\n- `Source code <https://github.com/otovo/python-netsgiro>`_\n- `Issue tracker <https://github.com/otovo/python-netsgiro/issues>`_\n',
    'author': 'Otovo AS',
    'author_email': 'jodal+netsgiro@otovo.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://pypi.org/project/netsgiro/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7',
}


setup(**setup_kwargs)
